from datetime import datetime

from rest_framework import routers, viewsets, filters
from rest_framework.permissions import IsAdminUser

from .models import Campaign
from .serializers import CampaignSerializer

from bulbs.utils.methods import get_query_params


class CampaignActiveFilter(filters.BaseFilterBackend):
    """Checks for a value for 'active' in query parameters, filters from this
    based on start_date and end_date. 'active' as True will return campaigns
    where start_date <= now < end_date and 'active' as False will return campaigns
    where now < start_date || now >= end_date."""

    def filter_queryset(self, request, queryset, view):

        new_queryset = queryset

        key_active = 'active'

        now = datetime.now()
        if key_active in get_query_params(request):
            val = get_query_params(request)[key_active].lower()
            if val == 'true':
                # start_date <= now < end_date
                new_queryset = queryset.filter(
                    start_date__lte=now,
                    end_date__gt=now
                )
            elif val == 'false':
                # now < start_date || now >= end_date
                new_queryset = \
                    queryset.filter(start_date__gt=now) | \
                    new_queryset.filter(end_date__lte=now)

        return new_queryset


class CampaignViewSet(viewsets.ModelViewSet):
    queryset = Campaign.objects.all()
    serializer_class = CampaignSerializer
    paginate_by = 10
    filter_backends = (
        CampaignActiveFilter,
        filters.SearchFilter,
        filters.OrderingFilter,)
    search_fields = (
        "campaign_label",
        "sponsor_name",)
    ordering_fields = (
        "campaign_label",
        "sponsor_name",
        "start_date",
        "end_date",)
    permission_classes = [IsAdminUser]


api_v1_router = routers.DefaultRouter()
api_v1_router.register(r"campaign", CampaignViewSet, base_name="campaign")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contributions', '0005_auto_20151130_1036'),
    ]

    operations = [
        migrations.AlterField(
            model_name='lineitem',
            name='amount',
            field=models.FloatField(default=0),
        ),
    ]



from django.conf import settings
from django.core.exceptions import ValidationError
from django.db.models.signals import pre_delete
from django.db import models
from django.template.defaultfilters import slugify

from elasticsearch import Elasticsearch
from json_field import JSONField

from bulbs.campaigns.models import Campaign
from bulbs.content.custom_search import custom_search_model
from bulbs.content.models import Content
from bulbs.content.mixins import DetailImageMixin
from bulbs.utils.methods import today_as_utc_datetime, is_valid_digit


es = Elasticsearch(settings.ES_URLS)


class SpecialCoverage(DetailImageMixin, models.Model):
    name = models.CharField(max_length=255, unique=True)
    slug = models.SlugField(max_length=255, blank=True, editable=True, unique=True)
    description = models.TextField(default="", blank=True)
    query = JSONField(default={}, blank=True)
    videos = JSONField(default=[], blank=True)
    active = models.BooleanField(default=False)
    promoted = models.BooleanField(default=False)
    start_date = models.DateTimeField(blank=True, null=True)
    end_date = models.DateTimeField(blank=True, null=True)
    campaign = models.ForeignKey(
        Campaign, null=True, default=None, blank=True, on_delete=models.SET_NULL)
    # Property-specific custom configuration
    config = JSONField(default={}, blank=True)

    def __unicode__(self):
        return self.name

    def clean(self):
        super(SpecialCoverage, self).clean()
        if self.query and self.query != {}:
            for key, value in self.query.items():
                if isinstance(value, list) and None in value:
                    self.query[key] = [v for v in value if v is not None]
        if self.videos:
            self.videos = [int(v) for v in self.videos if v is not None and is_valid_digit(v)]

    def save(self, *args, **kwargs):
        """Saving ensures that the slug, if not set, is set to the slugified name."""
        self.clean()
        self.validate_publish_dates()

        if not self.slug:
            self.slug = slugify(self.name)

        super(SpecialCoverage, self).save(*args, **kwargs)

        if self.query and self.query != {}:
            if self.is_active:
                self._save_percolator()
            else:
                self._delete_percolator()

    def validate_publish_dates(self):
        """
        If an end_date value is provided, the start_date must be less.
        """
        if self.end_date:
            if not self.start_date:
                raise ValidationError("""The End Date requires a Start Date value.""")
            elif self.end_date < self.start_date:
                raise ValidationError("""The End Date must not precede the Start Date.""")
        if self.start_date and not self.end_date:
            raise ValidationError("""The Start Date requires an End Date.""")

    def _save_percolator(self):
        """saves the query field as an elasticsearch percolator
        """
        index = Content.search_objects.mapping.index
        query_filter = self.get_content(published=False).to_dict()

        q = {}

        if "query" in query_filter:
            q = {"query": query_filter.get("query", {})}
        else:
            # We don't know how to save this
            return

        # We'll need this data, to decide which special coverage section to use
        if self.campaign:
            q["sponsored"] = True
            q["start_date"] = self.campaign.start_date
            q["end_date"] = self.campaign.end_date

        es.index(
            index=index,
            doc_type=".percolator",
            body=q,
            id=self.es_id
        )

    def _delete_percolator(self):
        index = Content.search_objects.mapping.index
        es.delete(index=index, doc_type=".percolator", id=self.es_id, refresh=True, ignore=404)

    def get_content(self, published=True):
        """performs es search and gets content objects
        """
        if "query" in self.query:
            q = self.query["query"]
        else:
            q = self.query
        search = custom_search_model(Content, q, published=published, field_map={
            "feature-type": "feature_type.slug",
            "tag": "tags.slug",
            "content-type": "_type"
        })
        return search

    @property
    def is_active(self):
        now = today_as_utc_datetime()
        if self.start_date and self.end_date:
            if self.start_date < now and self.end_date > now:
                return True
        return False

    @property
    def es_id(self):
        return "specialcoverage.{}".format(self.id)

    @property
    def contents(self):
        """performs .get_content() and caches it in a ._content property
        """
        if not hasattr(self, "_content"):
            self._content = self.get_content()
        return self._content

    @property
    def has_pinned_content(self):
        """determines if the there is a pinned object in the search
        """
        if "query" in self.query:
            q = self.query["query"]
        else:
            q = self.query
        if "pinned_ids" in q:
            return bool(len(q.get("pinned_ids", [])))
        return False


def remove_percolator(sender, instance, *args, **kwargs):
    instance._delete_percolator()

pre_delete.connect(remove_percolator, sender=SpecialCoverage)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django
from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('content', '0002_add_groups'),
    ]

    if django.VERSION >= (1, 8, 0):
        dependencies.insert(0,
            ('contenttypes', '0002_remove_content_type_name'))

    operations = [
        migrations.CreateModel(
            name='TemplateType',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=255)),
                ('slug', models.SlugField(unique=True)),
                ('content_type', models.ForeignKey(to='contenttypes.ContentType')),
            ],
        ),
        migrations.AddField(
            model_name='content',
            name='template_type',
            field=models.ForeignKey(blank=True, to='content.TemplateType', null=True),
        ),
    ]



from django.conf import settings
from django.contrib.auth import get_user_model

from .models import Contribution, ContributorRole, FreelanceProfile, FEATURETYPE


def get_forced_payment_contributions(start_date, end_date, qs=None):
    if not qs:
        qs = Contribution.objects.filter(force_payment=True)
    include = qs.filter(
        payment_date__range=(start_date, end_date)
    ) | qs.filter(
        payment_date__isnull=True
    )
    exclude = qs.exclude(
        payment_date__isnull=True
    ).exclude(
        payment_date__range=(start_date, end_date),
    )
    return include, exclude


def update_content_contributions(content, author):
    if not Contribution.objects.filter(content=content, contributor=author).exists():
        role_name = getattr(settings, 'DEFAULT_CONTRIBUTOR_ROLE', 'default')
        role, created = ContributorRole.objects.get_or_create(name=role_name)
        if created:
            role.payment_type = FEATURETYPE
            role.save()
        Contribution.objects.create(content=content, contributor=author, role=role)


def import_payroll_names(lookup_string):
    User = get_user_model()
    lookup_string = lookup_string.decode('unicode_escape').encode('ascii', 'ignore')
    rows = lookup_string.split('\n')
    rows = [row.rstrip('\n').rstrip('\r') for row in rows]
    rows = [row.split('\t') for row in rows]
    for row in rows:
        full_name, payroll_name = row[0], row[1]
        ordered_name = full_name.strip().split(' ')
        if len(ordered_name) == 2:
            first_name, last_name = ordered_name[0], ordered_name[1]
            qs = User.objects.filter(first_name=first_name, last_name=last_name)
            if qs.count() != 1:
                # print 'Cannot find user: {}'.format(full_name)
                return
            else:
                user = qs.first()
                profile = getattr(user, 'freelanceprofile', None)
                if not profile:
                    profile = FreelanceProfile.objects.create(
                        contributor=user,
                        is_freelance=True
                    )
                else:
                    profile.payroll_name = payroll_name.strip()
                    profile.save()



from django.contrib.auth import get_user_model
from django.conf import settings
from django.db import models

from elasticsearch_dsl import field
from djes.models import Indexable, IndexableManager

from bulbs.content.models import Content, FeatureType


FLAT_RATE = 0
FEATURETYPE = 1
HOURLY = 2
MANUAL = 3
OVERRIDE = 4

ROLE_PAYMENT_TYPES = (
    (FLAT_RATE, 'Flat Rate'),
    (FEATURETYPE, 'FeatureType'),
    (HOURLY, 'Hourly'),
    (MANUAL, 'Manual')
)

RATE_PAYMENT_TYPES = ROLE_PAYMENT_TYPES + ((OVERRIDE, 'Override'),)


def calculate_hourly_pay(rate, minutes_worked):
    return ((float(rate) / 60) * minutes_worked)


class SlugObjectField(field.Object):

    def __init__(self, *args, **kwargs):
        super(SlugObjectField, self).__init__(*args, **kwargs)
        self.properties['slug'] = field.construct_field('string', index='not_analyzed')


class ContentField(field.Object):

    def __init__(self, *args, **kwargs):
        super(ContentField, self).__init__(*args, **kwargs)
        self.properties['slug'] = field.construct_field('string', not_analyzed=True)
        self.properties['feature_type'] = SlugObjectField(not_analyzed=True)
        self.properties['tags'] = field.construct_field({
            'type': 'nested',
            'properties': {
                'slug': {'type': 'string', 'index': 'not_analyzed'}
            }
        })

    def to_es(self, obj):
        doc = {
            'id': obj.id,
            'title': obj.title,
            'published': obj.published,
            'feature_type': obj.feature_type.to_dict() if obj.feature_type else {}
        }

        return doc

    def to_python(self, data):
        return Content.objects.get(id=data['id'])


class ContributorField(field.Object):

    def __init__(self, *args, **kwargs):
        super(ContributorField, self).__init__(*args, **kwargs)
        self.properties['id'] = field.Long()
        self.properties['username'] = field.String(index='not_analyzed')
        self.properties['payroll_name'] = field.String(index='not_analyzed')
        self.properties['is_freelance'] = field.Boolean()

    def to_es(self, obj):
        data = {
            'id': obj.id,
            'username': obj.username
        }
        profile = getattr(obj, 'freelanceprofile', None)
        if profile:
            data['is_freelance'] = profile.is_freelance
            data['payroll_name'] = getattr(profile, 'payroll_name', '')
        return data

    def to_python(self, data):
        user_cls = get_user_model()
        user = user_cls.objects.filter(id=data['id'])
        if user.exists():
            return user.first()


class ContributionField(field.Object):

    def __init__(self, *args, **kwargs):
        super(ContributionField, self).__init__(*args, **kwargs)
        self.properties['id'] = field.Long()
        self.properties['contributor'] = ContributorField()

    def to_es(self, obj):
        if issubclass(obj.__class__, models.Manager):
            data = []
            for instance in obj.all():
                obj_data = {}
                for prop in self.properties:
                    c_field = self.properties[prop]
                    ref = getattr(instance, prop)
                    if hasattr(c_field, 'to_es'):
                        obj_data[prop] = c_field.to_es(ref)
                    elif hasattr(ref, 'to_dict'):
                        obj_data[prop] = ref.to_dict()
                    else:
                        obj_data[prop] = ref
                data.append(obj_data)
        else:
            data = {}
            for prop in self.properties:
                c_field = self.properties[prop]
                ref = getattr(obj, prop)
                if hasattr(c_field, 'to_es'):
                    data[prop] = c_field.to_es(ref)
                elif hasattr(ref, 'to_dict'):
                    data[prop] = ref.to_dict()
                else:
                    data[prop] = ref
        return data

    def to_python(self, data):
        return data


class ReportContent(Content):

    reference = models.Manager()

    class Meta:
        proxy = True

    class Mapping(Content.Mapping):

        # contributions = ContributionField()
        tags = field.Nested()

        class Meta:
            orphaned = True
            includes = ('tags',)
            # excludes = ('contributions',)


class LineItem(models.Model):
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL)
    amount = models.FloatField(default=0)
    note = models.TextField()
    payment_date = models.DateTimeField(null=True, blank=True)


class ContributorRole(Indexable):
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, blank=True)
    payment_type = models.IntegerField(choices=ROLE_PAYMENT_TYPES, default=MANUAL)

    def save(self, *args, **kwargs):
        created = bool(self.pk is None)
        super(ContributorRole, self).save(*args, **kwargs)
        self.create_feature_type_rates(created)

    def create_feature_type_rates(self, created=False):
        """
        If the role is being created we want to populate a rate for all existing feature_types.
        """
        if created:
            for feature_type in FeatureType.objects.all():
                FeatureTypeRate.objects.create(role=self, feature_type=feature_type, rate=0)

    def get_rate(self):
        if self.payment_type == FLAT_RATE:
            qs = self.flat_rates.all()
            if qs.exists():
                return qs.first()
        if self.payment_type == HOURLY:
            qs = self.hourly_rates.all()
            if qs.exists():
                return qs.first()
        return None


class Contribution(Indexable):
    role = models.ForeignKey(ContributorRole)
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL, related_name="contributions")
    content = models.ForeignKey(Content, related_name="contributions")
    notes = models.TextField(null=True, blank=True)
    minutes_worked = models.IntegerField(null=True)
    force_payment = models.BooleanField(default=False)
    payment_date = models.DateTimeField(null=True, blank=True)

    search_objects = IndexableManager()

    class Mapping:

        content = ContentField()
        contributor = ContributorField()
        pay = field.Long()

        class Meta:
            dynamic = False
            excludes = ('content', 'contributor')

    @property
    def pay(self):
        return self.get_pay

    @property
    def get_pay(self):
        return self._get_pay()

    @property
    def get_override(self):
        return self._get_override()

    def delete(self):
        self.delete_index()
        return super(Contribution, self).delete()

    def get_rate(self):
        payment_type = self.role.payment_type
        if payment_type == MANUAL:
            return self.manual_rates.all().first()

        if payment_type == FLAT_RATE:
            return self.role.flat_rates.filter().first()

        if self.content.feature_type and payment_type == FEATURETYPE:
            return self.content.feature_type.feature_type_rates.filter(
                role=self.role
            ).first()

        if payment_type == HOURLY:
            return self.role.hourly_rates.filter().first()

    def _get_override(self):
        # Get contribution specific overrides first.
        if self.role.payment_type in [0, 1] and self.override_contribution.exists():
            return self.override_contribution.first().rate

        override_profile = self.contributor.overrides.filter(role=self.role).first()
        if not override_profile:
            return None

        elif self.role.payment_type == 0 and override_profile.override_flatrate.exists():
            return override_profile.override_flatrate.first().rate

        elif (self.role.payment_type == 1):
            override = override_profile.override_feature_type.filter(
                feature_type=self.content.feature_type
            ).first()
            if override:
                return override.rate

        elif self.role.payment_type == 2 and override_profile.override_hourly.exists():
            minutes_worked = float(getattr(self, 'minutes_worked', 0))
            return calculate_hourly_pay(
                override_profile.override_hourly.first().rate, minutes_worked
            )

    def _get_pay(self):
        override = self.get_override
        if override:
            return override
        rate = self.get_rate()
        if isinstance(rate, HourlyRate):
            minutes_worked = getattr(self, 'minutes_worked', 0)
            if minutes_worked is None:
                minutes_worked = 0
            return calculate_hourly_pay(rate.rate, float(minutes_worked))
        if rate:
            return rate.rate
        return None


class OverrideProfile(Indexable):
    contributor = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='overrides')
    role = models.ForeignKey(ContributorRole, related_name='overrides')

    class Meta:
        unique_together = (('contributor', 'role'),)


class BaseOverride(Indexable):
    rate = models.IntegerField(default=0)
    updated_on = models.DateTimeField(auto_now_add=True)

    class Meta:
        ordering = ('-updated_on',)


class FeatureTypeOverride(BaseOverride):
    profile = models.ForeignKey(OverrideProfile, related_name='override_feature_type')
    feature_type = models.ForeignKey(FeatureType)


class FlatRateOverride(BaseOverride):
    profile = models.ForeignKey(OverrideProfile, related_name='override_flatrate')


class HourlyOverride(BaseOverride):
    profile = models.ForeignKey(OverrideProfile, related_name='override_hourly')


class ContributionOverride(BaseOverride):
    contribution = models.ForeignKey(Contribution, related_name='override_contribution')

    class Mapping:
        contribution = ContributionField()


class Rate(Indexable):
    name = models.IntegerField(choices=RATE_PAYMENT_TYPES, null=True)
    updated_on = models.DateTimeField(auto_now=True)
    rate = models.IntegerField()

    class Meta:
        ordering = ('-updated_on',)


class FlatRate(Rate):
    role = models.ForeignKey(ContributorRole, related_name="flat_rates")


class HourlyRate(Rate):
    role = models.ForeignKey(ContributorRole, related_name="hourly_rates")


class ManualRate(Rate):
    contribution = models.ForeignKey(Contribution, related_name="manual_rates")

    class Mapping:
        contribution = ContributionField()


class FeatureTypeRate(Rate):
    role = models.ForeignKey(ContributorRole, null=True, related_name="feature_type_rates")
    feature_type = models.ForeignKey(FeatureType, related_name="feature_type_rates")

    class Meta:
        unique_together = (("role", "feature_type"))
        ordering = ("feature_type__name",)

    @property
    def feature_type_slug(self):
        return getattr(self.feature_type, "slug", None)


class FreelanceProfile(Indexable):
    payroll_name = models.CharField(max_length=256, blank=True, null=True)
    contributor = models.OneToOneField(settings.AUTH_USER_MODEL, unique=True)
    is_freelance = models.BooleanField(default=True)
    is_manager = models.BooleanField(default=False)
    payment_date = models.DateTimeField(null=True, blank=True)

    def get_pay(self, start=None, end=None):
        qs = self.contributor.contributions.all()
        if start:
            qs = qs.filter(payment_date__gte=start)
        if end:
            qs = qs.filter(payment_date__lte=end)

        pay = 0
        for contribution in qs.all():
            contribution_pay = contribution.get_pay
            if contribution_pay:
                pay += contribution_pay
        return pay



from django.db import models

from djbetty.fields import ImageField
from djbetty.serializers import ImageFieldSerializer

from rest_framework import serializers


class BodyMixin(models.Model):
    """Provides a "body" field which contains HTML but is not included in the ES index body."""
    body = models.TextField(blank=True, default="")

    class Meta:
        abstract = True

class DetailImageMixin(models.Model):
    """Provides an "image" field, with caption and alt."""
    _image_caption = models.CharField(null=True, blank=True, editable=False, max_length=255)
    _image_alt = models.CharField(null=True, blank=True, editable=False, max_length=255)
    image = ImageField(null=True, blank=True, alt_field="_image_alt", caption_field="_image_caption")

    class Meta:
        abstract = True

class ImageSerializerMixin(serializers.Serializer):

    image = ImageFieldSerializer(required=False, allow_null=True)

    class Meta:
        abstract = True




CUSTOM_SEARCH_TIME_PERIODS = [{
    "label": "Past day",
    "days": 1
}, {
    "label": "Past week",
    "days": 7
}, {
    "label": "Past 2 weeks",
    "days": 14
}, {
    "label": "Past month",
    "days": 31
}, {
    "label": "Past 6 months",
    "days": 31 * 6
}, {
    "label": "Past year",
    "days": 365
}]



from rest_framework import serializers


class JSONField(serializers.Field):
    """Ensures json dict stays as a dict from db to json and back."""

    def to_representation(self, obj):
        if obj is None:
            return self.default
        return obj

    def to_internal_value(self, data):
        if data is None:
            return self.default
        return data
